

set -ex



python -m pip check || true
rm -f test/test_models.py
pytest --verbose -k "not (test_url_is_accessible or test_classes or test_feature_types or test_feature_types or test_is_valid_file or test_num_examples or test_smoke or test_str_smoke or test_transforms or test_adjust_gamma or test_forward_backward or test_jit_forward_backward)" test/
exit 0
